//
// Mars.java
// Mars calculations.
// Copyright 2003 by Akkana Peck,
// adapting some code from Marsmap, by Akkana Peck and Brian Tung.
// You are free to use this under the terms of the GPL.
//

class Planet
{
    static final double DEGREES = Math.PI/180.0;

    double inc;           /* inclination of orbit (in degrees) */
    double inc_b;
    double inc_m;
    double asc;           /* longitude of ascending node (in degrees) */
    double asc_b;
    double asc_m;
    double per;           /* longitude of perihelion (in degrees) */
    double per_b;
    double per_m;
    double sem;           /* semi-major axis (in AU) */
    double sem_b;
    double sem_m;
    double ecc;           /* eccentricity of orbit */
    double ecc_b;
    double ecc_m;
    double lon;           /* argument of longitude (in degrees) */
    double lon_b;
    double lon_m;

    Planet(double ib, double im,
           double ab, double am,
           double pb, double pm,
           double sb, double sm,
           double eb, double em,
           double lb, double lm)
    {
        inc_b = ib;  inc_m = im;
        asc_b = ab;  asc_m = am;
        per_b = pb;  per_m = pm;
        sem_b = sb;  sem_m = sm;
        ecc_b = eb;  ecc_m = em;
        lon_b = lb;  lon_m = lm;
    }

    public void ProcElements(double jd)
    {
        inc = (inc_b+jd*inc_m)*DEGREES;
        asc = (asc_b+jd*asc_m)*DEGREES;
        per = (per_b+jd*per_m)*DEGREES;
        sem = (sem_b+jd*sem_m);
        ecc = (ecc_b+jd*ecc_m);
        lon = (lon_b+jd*lon_m)*DEGREES;
    }
}

public class Mars
{
    /* XXX: these are just guesses; we should improve this */
    static final double northLon = -6.2;   /* longitude of north pole (deg) */
    static final double northObl = 26.2;   /* Mars obliquity (in degrees) */
    static final double lonCorr = 0.174;   /* correction for ref longitude */
    static final double marsRot = 1.025972; /* rot period in Earth days */
    Planet earth;
    Planet mars;

    // Values calculated by CalcCM:
    double CM;    // central meridian longitude
    double lat;   // latitude of center of disk
    double size;  // size of disk
    double orient; // angle of disk
    double dist;  // distance from earth

    public double getCM()     { return CM; }
    public double getLat()    { return lat; }
    public double getSize()   { return size; }
    public double getOrient() { return orient; }
    public double getDist()   { return dist; }

    static final double DEGREES = Math.PI/180.0;

    public Mars()
    {
        earth = new Planet(0.00005, -0.000000356985,
                           -11.26084, 0.0000000000,
                           102.9404, 0.0000470935,
                           1.00000011, -0.00000000000136893,
                           0.016709, -0.000000001151,
                           98.9874, 0.9856473520);
        mars = new Planet(1.8497, -0.0000000178,
                          49.5574, 0.0000211081,
                          336.0590, 0.0000504042,
                          1.52366231, -0.000000001977,
                          0.093405, 0.000000002516,
                          354.6611, 0.5240711808);
    }

    void CalcCM(double jdate)
    {
        double earthM, earthV, earthR;
        double marsM, marsV, marsR;
        double dotQR;
        double dotME;

        // But the calculation code actually doesn't want Julian date,
        // but something else, so convert to what it wants:
        jdate -= 2451543.497656;
        //System.out.println("jdate = " + new Double(jdate));

        earth.ProcElements(jdate);
        mars.ProcElements(jdate);

        /* compute ecliptic coordinates of the Earth */
        earthM = earth.lon-earth.per;
        earthV = Kepler(earthM, earth.ecc);
        earthR = earth.sem*(1.0-earth.ecc*earth.ecc)
            /(1.0+earth.ecc*FastTrig.cos(earthV));
        SimpleVector earthPos = new SimpleVector(
            earthR*FastTrig.cos(earthV+earth.per),
            earthR*FastTrig.sin(earthV+earth.per),
            0.0);

        /* compute ecliptic coordinates of Mars */
        marsM = mars.lon-mars.per;
        marsV = Kepler(marsM, mars.ecc);
        marsR = mars.sem*(1.0-mars.ecc*mars.ecc)
            /(1.0+mars.ecc*FastTrig.cos(marsV));

        SimpleVector marsPos = new SimpleVector(
            marsR*(FastTrig.cos(mars.asc)*FastTrig.cos(marsV+mars.per-mars.asc)-
                   FastTrig.sin(mars.asc)*FastTrig.sin(marsV+mars.per-mars.asc)
                   *FastTrig.cos(mars.inc)),
            marsR*(FastTrig.sin(mars.asc)*FastTrig.cos(marsV+mars.per-mars.asc)
                   + FastTrig.cos(mars.asc)*FastTrig.sin(marsV+mars.per-mars.asc)
                   *FastTrig.cos(mars.inc)),
            marsR*FastTrig.sin(marsV+mars.per-mars.asc)*FastTrig.sin(mars.inc));

        /* compute vector from center of Mars to center of Earth */
        SimpleVector mars2Earth = new SimpleVector (
            earthPos.x-marsPos.x,
            earthPos.y-marsPos.y,
            earthPos.z-marsPos.z);
        dist = Math.sqrt(mars2Earth.dot(mars2Earth));
        mars2Earth.NormalizeVector();
        /* note: centerXYZ also points from center of Mars to center of disc */

        /* compute vector from center of Mars to Mars's north pole */
        SimpleVector marsNorth = new SimpleVector (
            FastTrig.sin(northObl*DEGREES)*FastTrig.cos(northLon*DEGREES),
            FastTrig.sin(northObl*DEGREES)*FastTrig.sin(northLon*DEGREES),
            FastTrig.cos(northObl*DEGREES));

        /* compute vector from center of Mars to equator of Mars
           at reference longitude */
        SimpleVector marsRef = new SimpleVector (
            FastTrig.cos(northObl*DEGREES)*FastTrig.cos(northLon*DEGREES),
            FastTrig.cos(northObl*DEGREES)*FastTrig.sin(northLon*DEGREES),
            -FastTrig.sin(northObl*DEGREES));
        /* note: the reference longitude is the Mars longitude of the
           ecliptic *south* pole */

        /* compute vector from center of Mars to equator of Mars
           at same longitude as center of disc */
        SimpleVector marsEq = mars2Earth;
        marsEq.OrthogonalizeVector(marsNorth);

        /* compute longitude of Mars at reference point */
        CM = ((jdate+lonCorr-dist*0.0057) % marsRot)
                        // XXX hope this does fmod
                   *2.0*Math.PI/marsRot;

        /* compute longitude of center of disc (CM longitude) */
        dotQR = marsEq.dot(marsRef);
        SimpleVector eqRefCross = marsEq.cross(marsRef);
        if (eqRefCross.z > 0.0)
            CM += Math.acos(dotQR);
        else
            CM -= Math.acos(dotQR);
        while (CM < 0.0)
            CM += 2.0*Math.PI;
        while (CM >= 2.0*Math.PI)
            CM -= 2.0*Math.PI;

        /* compute latitude of center of disc */
//        System.out.println("mars2Earth: " + mars2Earth.toString());
//        System.out.println("marsNorth: " + marsNorth.toString());
//        System.out.println("0.5*PI - acos "
//                           + new Double(mars2Earth.dot(marsNorth)));
//        System.out.println("= 0.5*PI - "
//                           + new Double(Math.acos(mars2Earth.dot(marsNorth))));
        lat = 0.5*Math.PI - Math.acos(mars2Earth.dot(marsNorth));

        /* compute size of disc */
        size = 9.37/dist;

        /* compute vector to CNP */
        SimpleVector earthNorth = new SimpleVector (
            0.0, FastTrig.sin(23.4*DEGREES), FastTrig.cos(23.4*DEGREES));

        /* project onto two-dimensional view of Mars */
        earthNorth.OrthogonalizeVector(mars2Earth);

        /* project north pole onto same view */
        marsNorth.OrthogonalizeVector(mars2Earth);

        /* compute orientation angle */
        dotME = marsNorth.dot(earthNorth);
        orient = Math.acos(dotME);
        SimpleVector marsEarthCross = marsNorth.cross(earthNorth);
        if (marsEarthCross.dot(mars2Earth) > 0.0)
            orient = -orient;

        //System.out.println("CM = " + new Integer((int)(CM * 360 / Math.PI))
        //                   + ", lat = "
        //                   + new Integer((int)(lat * 360 / Math.PI))
        //                   + ", size = "
        //                   + new Double(size));
    }

    double Kepler(double mean, double ecc)
    {
        double e;
        double delta;

        while (mean < 0.0)
            mean += 2.0*Math.PI;
        while (mean >= 2.0*Math.PI)
            mean -= 2.0*Math.PI;
        e = mean;
        delta = 0.05;
        while (Math.abs(delta) >= 0.00001) {
            delta = e-ecc*FastTrig.sin(e)-mean;
            e -= delta/(1.0-ecc*FastTrig.cos(e));
        }
        return(2.0*Math.atan(Math.sqrt((1.0+ecc)/(1.0-ecc))*Math.tan(0.5*e)));
    }
}


