//
// SphereMap.java
// Map a rectangular map onto a sphere.
// Copyright 2003 by Akkana Peck.
// You are free to use this under the terms of the GPL.
//

//
// SphereMap assumes that its cylindrical map argument
// has 0 meridian at the center, and north up.
// CM and Latitude are specified in degrees.
//

import java.awt.*;
import java.awt.image.*;
import java.util.*;

public class SphereMap
{
    static final double PI_2 = Math.PI / 2.;

public Image mapImageToSphere(Image mapImage, Component comp,
                              int cm, int lat,
                              boolean northUp,
                              boolean eastRight)
    {
        // Grab input image into pixels_color array using PixelGrabber
        // Initialize color arrays
        int w = mapImage.getWidth(comp);
        int h = mapImage.getHeight(comp);
        int newH = h;
        int newW = (int)((double)w / PI_2);
        newW = newH;
//        System.out.println("Old: " + new Integer(w).toString()
//                           + " x " + new Integer(h).toString());
//        System.out.println("New: " + new Integer(newW).toString()
//                           + " x " + new Integer(newH).toString());
//        System.out.println("lat: " + new Integer(lat));
        if (w < 0 || h < 0) {
            return null;
        }
        if (!eastRight) cm = -cm;
        int cmIm = cm * w / 360 + w/2;
        if (cmIm < 0) cmIm += w;
        int[] mapPxls = new int[h * w];
        PixelGrabber grabber = new PixelGrabber(mapImage, 0, 0, w, h,
                                                mapPxls, 0, w);
        try {
            grabber.grabPixels();
        } catch (InterruptedException e) {
            System.out.println("grabber error" + e);
        }
        if ((grabber.status() & ImageObserver.ABORT) != 0)
            System.out.println("grabber error");

        // Make new image buffer
        int[] newPxls = new int[newH * newW];

        // Find the limits of visibility
//        System.out.println("cmIm: " + new Integer(cmIm));
        int xmin = (cmIm - w/4);
        if (xmin < 0) xmin += w;
        int xmax = (cmIm + w/4) % w;
        int x = xmin;
        int xcenter = (xmin + xmax) / 2;
        SimpleVector up = new SimpleVector(0., 0., 1.);
        SimpleVector right = new SimpleVector(1., 0., 0.);
//        System.out.println("newWidth: " + new Double(newW));
//        System.out.println("xmin: " + new Integer(xmin) + ", xmax: "
//                           + new Integer(xmax));
        while (true)  // Loop over x -- boundary conditions inside
        {
            for (int y = 0; y < h; y++)
            {
                // Convert (x, y) to latitude and longitude on the sphere.
                double latitude = Math.PI * (.5 - (h/2. - y) / h - lat / 180.);
                double longitude = Math.PI * ((x * 2.) / w - 1. - cm / 180.);

                // Vector math
                double latplus = latitude + PI_2;
                double longplus = longitude + PI_2;
                SimpleVector vect = new SimpleVector(
                    Math.cos(latplus)*Math.cos(longplus),
                    -Math.cos(latplus)
                     *Math.sin(longplus),
                    Math.sin(latplus));
                //if (DotProduct(vect, cent) < 0.0)
                //    return;

                double xx = newW/2 * (1 + vect.dot(right));
                double yy = newH/2 * (1 - vect.dot(up));

                // Fill in "holes" that show up:
                try {
                    int idx = (int)xx + ((int)yy * newW);
                    int val = mapPxls[x + (y * w)];
                    newPxls[idx] = val;
                    newPxls[idx+1] = val;
                    newPxls[idx+newW] = val;
                    newPxls[idx+newW+1] = val;
                }
                catch (ArrayIndexOutOfBoundsException e) {
//                  System.out.println("(" + new Integer((int)x) + ", " +
//                                     new Integer((int)y) + "); " +
//                                     "(" +
//                                     new Integer((int)(longitude*180/Math.PI))
//                                     + ", " +
//                                     new Integer((int)(latitude*180/Math.PI))
//                                     + "); " +
//                                     "(" + new Integer((int)xx) + ", " +
//                                     new Integer((int)yy) + ")");
                }
            }
            // Boundary conditions:
            ++x;
            if (xmax > xmin) {
//                if (x % 10 == 0)
//                    System.out.println("non-wrapping: x = " + new Integer(x));
                if (x >= xmax) break;
                else continue;
            }
            // Else xmax < xmin and we have to wrap around.
            // Have we already wrapped around?
            if (x < xmin) {
//                if (x % 10 == 0)
//                    System.out.println("wrapped: x = " + new Integer(x));
                if (x >= xmax) break;
                else continue;
            }
            // We haven't wrapped yet.
            if (x >= w) {     // Time to wrap!
//                System.out.println("Wrapping!");
                x = 0;
                continue;
            }
//            if (x % 10 == 0)
//                System.out.println("pre-wrapping: x = " + new Integer(x));
        }

        // Create image out of buffer array
        Image globeImage = comp.createImage
            (new MemoryImageSource(newW, newH, newPxls, 0, newW));

        MediaTracker mt = new MediaTracker(comp);
        mt.addImage(globeImage, 0);
        try {
            mt.waitForID(0);
        } catch (InterruptedException e) {
            System.out.println("Media Tracker Error " + e);
        }

        return globeImage;
    }
}

